import copy
import inspect
import logging
import types
from typing import Callable, Optional, ParamSpec, TypeVar
import lastmile_utils.lib.core.utils as utils

from result import Err, Ok, Result

from lastmile_utils.lib.core.utils import InternalError

T_ParamSpec = ParamSpec("T_ParamSpec")
T_cov = TypeVar("T_cov", covariant=True)
T_Exn = TypeVar("T_Exn", bound=Exception)
EResult = Result[T_cov, Exception]

logger = logging.getLogger(__name__)

# TODO configure level
logging.basicConfig(level=logging.INFO, format=utils.LOGGER_FMT)


def resultify_exceptions(
    f: Callable[T_ParamSpec, T_cov]
) -> Callable[T_ParamSpec, EResult[T_cov]]:
    """
    Turn a callable that might raise an exception into a callable that
    returns a Result type.
    This also stores convenient traceback information.

    Generally, you should use this when calling out across code boundaries
    into 3rd party code (that can raise Exceptions).

    That is, use this at the lowest possible level.

    This can also be used for user-provided callbacks.

    Output: a function that is the same as the input function,
    but where the output / any Exception is wrapped in a EResult
        (with lots of convenient traceback info).
    """

    def inner(
        *args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs
    ) -> EResult[T_cov]:
        try:
            return Ok(f(*args, **kwargs))
        except Exception as e:
            frame = inspect.currentframe()
            e_new = copy.copy(e)
            tb_to_here = _frame_to_tb(frame)
            e_new.__traceback__ = tb_to_here
            e_new.__cause__ = e
            return Err(e_new)

    return inner


def make_err(e: Exception) -> EResult[T_cov]:
    """
    Use this as an alternative to the Err constructor for nicer
    traceback info.
    """

    @resultify_exceptions
    def _inner() -> T_cov:
        raise e

    return _inner()


def result_return_or_raise_for_apis_only(
    fn: Callable[T_ParamSpec, EResult[T_cov]]
) -> Callable[T_ParamSpec, T_cov]:
    """
    PLEASE ONLY USE THIS AT THE TOP LEVEL OF YOUR APPLICATION,
    AT USER INTERFACES.

    **Description**

    The purpose of this is to provide users with the standard
    Python exception outcome, namely, raising an exception.

    You should use it at LM API boundaries and probably nowhere else.

    Because this function raises, it should be used judiciously.
    It is intentionally named to deter usage, or at least to provide
    something to grep for during debugging.



    **Input / Output**

    It essentially inverts `resultify_exceptions()`. That is,
    it takes a function returning a EResult, and outputs an analogous function
    that returns the inside value type, or raises an exception.

    It also does some convenient logging and handling of
    different types of exceptions.


    See examples/library_structure/my_library/api.py for an example.
    """

    def return_or_raise_(
        *args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs
    ) -> T_cov:
        try:
            result = fn(*args, **kwargs)
        except Exception as e:
            message = f"""
            Unexpected error!
            An Exception was caught inside 
            result_return_or_raise_for_apis_only().
            
            This means there is probably a missing application of the 
            resultify_exceptions() decorator. Please see the stack trace.

            Please file a bug report...
            """
            logger.critical(message)
            raise InternalError(message) from e

        match result:
            case Ok(value):
                return value
            case Err(e):
                logger.error(f"Exception: {e}")
                raise e

    return return_or_raise_


def _frame_to_tb(
    frame: Optional[types.FrameType],
) -> Optional[types.TracebackType]:
    """This stuff really brings me back to freshman CS"""
    return _frame_to_tb_rec(frame, None)


def _frame_to_tb_rec(
    frame: Optional[types.FrameType], acc: Optional[types.TracebackType]
) -> Optional[types.TracebackType]:
    """
    This one especially brings up nostalgia for
    [CS19](https://cs.brown.edu/courses/cs019/).

    I'm almost tempted to write this in Scheme...
    """
    if frame is None:
        return acc
    else:
        return _frame_to_tb_rec(
            frame.f_back,
            types.TracebackType(acc, frame, frame.f_lasti, frame.f_lineno),
        )



import argparse
from dataclasses import dataclass
import logging
import os
from enum import Enum, EnumMeta
from types import UnionType
from typing import Any, Dict, Mapping, Optional, Set, Type, TypeVar
import typing

import pydantic
from dotenv import load_dotenv
from result import Ok, Result

from .utils import (
    ErrWithTraceback,
    JSONObject,
    Record,
    dict_union_allow_replace,
    remove_nones,
)

LOGGER = logging.getLogger(__name__)


@dataclass
class TypeOrigin:
    origin: type
    args: list["TypeOrigin"]


def _alias_type_to_origin_type(
    some_type: Type[Any] | type,
) -> TypeOrigin | None:
    args = (
        []
        if not hasattr(some_type, "__args__")
        else [
            _alias_type_to_origin_type(a)
            for a in getattr(some_type, "__args__")
        ]
    )
    args = [a for a in args if a is not None]
    constructor = getattr(some_type, "__origin__", some_type)
    return TypeOrigin(constructor, args)


def _get_inside_of_optional_type(
    some_type: Type[Any] | type,
) -> Optional[type]:
    """
    examples:
    - Optional[int] -> int
    - Optional[Optional[list[str]]] -> list
    - list[str] -> None
    """
    origin_type = _alias_type_to_origin_type(some_type)
    if origin_type is not None and origin_type.origin == typing.Union:
        type_args = getattr(some_type, "__args__")
        if not (len(type_args) == 2 and type_args[1] == type(None)):
            return None

        inside_type = type_args[0]
        if not hasattr(inside_type, "__origin__"):
            return inside_type
        else:
            return getattr(inside_type, "__origin__")


def _parse_log_level(x: int | str) -> int:
    match x:
        case int(x_i):
            return x_i
        case str(x_s):
            out = logging.getLevelName(x_s.strip().upper())
            match out:
                case int(out_i):
                    return out_i
                case _:
                    msg = "Choices: ['CRITICAL', 'FATAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']"
                    raise argparse.ArgumentTypeError(msg)


def add_parser_argument(
    parser: argparse.ArgumentParser,
    field_name: str,
    field: pydantic.fields.FieldInfo,
    is_required: bool,
) -> Optional[str]:
    return add_parser_argument_helper(
        parser, field_name, field.annotation, is_required
    )


def add_parser_argument_helper(
    parser: argparse.ArgumentParser,
    field_name: str,
    field_type: Type[Any] | type | None,
    is_required: bool,
) -> Optional[str]:
    field_name = field_name.replace("_", "-")
    inside_of_optional_type = (
        _get_inside_of_optional_type(field_type)
        if field_type is not None
        else None
    )
    if inside_of_optional_type is not None:
        return add_parser_argument_helper(
            parser, field_name, inside_of_optional_type, False
        )
    elif field_name in ["log-level"]:
        parser.add_argument(
            f"--{field_name}",
            type=_parse_log_level,
            required=is_required,
        )
    elif field_type is None:
        return f"{field_name}: type is None"
    elif field_type is bool:
        if is_required:
            return f"(bool) flag cannot be required: field={field_name}"
        parser.add_argument(
            f"--{field_name}",
            action="store_true",
        )
    elif type(field_type) is UnionType:
        return f"UnionType not supported. If your field is a `type | None`, use Optional[type] instead. field={field_name}"
    elif isinstance(field_type, EnumMeta):
        parser.add_argument(
            f"--{field_name}",
            type=str,
            choices=[e.value.lower() for e in field_type],  # type: ignore
            required=is_required,
        )
    elif field_type is list:
        parser.add_argument(
            f"--{field_name}",
            nargs="+",
            required=is_required,
        )
    else:
        parser.add_argument(
            f"--{field_name}", type=field_type, required=is_required
        )


def add_parser_arguments(
    parser: argparse.ArgumentParser,
    fields: dict[str, pydantic.fields.FieldInfo],
    required: Optional[Set[str]] = None,
):
    required = required or {
        f_name for f_name, f in fields.items() if f.is_required()
    }
    for field_name, field in fields.items():
        is_required = field_name in required
        res = add_parser_argument(parser, field_name, field, is_required)
        if res is not None:
            LOGGER.warning(res)


def argparsify(
    r: Record | Type[Record],
    required: Optional[Set[str]] = None,
    subparser_record_types: Mapping[str, Record | Type[Record]] = {},
) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    add_parser_arguments(parser, r.model_fields, required=required)

    if subparser_record_types:
        subparsers = parser.add_subparsers(dest="subcommand", required=True)
        for subparser_name, subparser_r in subparser_record_types.items():
            subparser = subparsers.add_parser(subparser_name)
            add_parser_arguments(subparser, subparser_r.model_fields)

    return parser


def get_config_args(
    args: argparse.Namespace | Dict[str, Any]
) -> Dict[str, Any]:
    def _get_cli():
        if isinstance(args, argparse.Namespace):
            return vars(args)
        else:
            return dict(args)

    env_keys = {
        "openai_key": "OPENAI_API_KEY",
        "pinecone_key": "PINECONE_API_KEY",
        "pinecone_environment": "PINECONE_ENVIRONMENT",
        "pinecone_index_name": "PINECONE_INDEX_NAME",
        "pinecone_namespace": "PINECONE_NAMESPACE",
    }

    load_dotenv()
    env_values = remove_nones({k: os.getenv(v) for k, v in env_keys.items()})
    cli_values = remove_nones(_get_cli())

    all_args = dict_union_allow_replace(
        env_values, cli_values, on_conflict="replace"
    )
    return all_args


def resolve_path(data_root: str, path: str) -> str:
    """
    data_root: path relative to cwd where all the data is stored.
    path: relative to data_root, the specific data to look at,
    e.g. "10ks/".
    """

    joined = os.path.join(os.getcwd(), data_root, path)
    return os.path.abspath(os.path.expanduser(joined))


T_Record = TypeVar("T_Record", bound=Record)


def get_subparser_name(
    main_parser: argparse.ArgumentParser, argv: list[str]
) -> str:
    parsed = main_parser.parse_args(argv)
    return parsed.subcommand


def parse_args(
    parser: argparse.ArgumentParser,
    argv: list[str],
    config_type: Type[T_Record],
) -> Result[T_Record, str]:
    parsed = parser.parse_args(argv)
    cli_dict = remove_nones(vars(parsed))
    return config_from_primitives(config_type, cli_dict)


def config_from_primitives(
    config_type: Type[T_Record], cli_dict: JSONObject
) -> Result[T_Record, str]:
    try:
        return Ok(config_type.model_validate(cli_dict))
    except pydantic.ValidationError as e:
        return ErrWithTraceback(e, "Invalid args")


T_Enum = TypeVar("T_Enum", bound=Enum)


def validate_enum_field_value(value: Any, field_type: Type[T_Enum]) -> T_Enum:
    if isinstance(value, str):
        try:
            out = field_type[value.upper()]
            return out
        except KeyError as e:
            raise ValueError(
                f"Unexpected value for {field_type.__name__}: {value}"
            ) from e
    elif not isinstance(value, field_type):
        raise ValueError(
            f"Unexpected type for {field_type.__name__}: {value}, {type(value)}"
        )
    else:
        return value


class EnumValidatedRecordMixin(pydantic.BaseModel):
    @pydantic.root_validator(pre=True)
    def check_enum_fields(cls, values: dict[str, Any]) -> dict[str, Any]:
        out: dict[str, Any] = {}
        for field_name, value in values.items():
            if field_name not in cls.model_fields:
                # This can happen if an extra field is given.
                # In this case, if all the other fields are valid,
                # Pydantic actually allows this and ignores the extra field.
                # Here, we're just checking Enum fields, so we can ignore this
                # field and continue to validate the other fields.
                continue
            field: pydantic.fields.FieldInfo = cls.model_fields[field_name]
            _type = field.annotation
            if isinstance(_type, EnumMeta):
                out[field_name] = validate_enum_field_value(value, _type)
            else:
                out[field_name] = value

        return out



"""Convenience module for importing from multiple libraries:
    `import lastmile_utils.lib.core.api as cu`

"""
# pylint: disable=wildcard-import,unused-wildcard-import
# pyright: reportWildcardImportFromLibrary=false



import asyncio
import traceback
from abc import abstractmethod
from typing import (
    Awaitable,
    Iterable,
    Mapping,
    ParamSpec,
    Protocol,
    Tuple,
    TypeVar,
)

from typing import Callable, Concatenate, ParamSpec, TypeVar
from result import Result, Ok


from result import Err, Ok, Result

# Types


TR = TypeVar("TR", covariant=True)
E = TypeVar("E", covariant=True)

K = TypeVar("K")
T = TypeVar("T")

PS = ParamSpec("PS")  # Capture the parameters
PS2 = ParamSpec("PS2")  # Capture the parameters
PS3 = ParamSpec("PS3")  # Capture the parameters
T_Return_cov = TypeVar(
    "T_Return_cov", covariant=True
)  # constrained on protocol call definition
T_Return_cov2 = TypeVar("T_Return_cov2", covariant=True)
T_Return_contra = TypeVar("T_Return_contra", contravariant=True)
T_Return = TypeVar("T_Return")
T_Return2 = TypeVar("T_Return2")
T_Err = TypeVar("T_Err")


class T_UnsafeFn(Protocol[PS, T_Return_cov]):
    """function type returning an arbitrary value"""

    @abstractmethod
    def __call__(self, *args: PS.args, **kwargs: PS.kwargs) -> T_Return_cov:
        pass


class T_SafeFn(Protocol[PS, T_Return_cov, E]):
    """function type returning a Result"""

    @abstractmethod
    def __call__(self, *args: PS.args, **kwargs: PS.kwargs) -> Result[T_Return_cov, E]:
        pass


class T_AsyncUnsafeFn(Protocol[PS, T_Return_cov]):
    """Async function type returning an arbitrary value"""

    @abstractmethod
    async def __call__(self, *args: PS.args, **kwargs: PS.kwargs) -> T_Return_cov:
        pass


class T_AsyncSafeFn(Protocol[PS, T_Return_cov, E]):
    """Async function type returning a Result"""

    @abstractmethod
    async def __call__(
        self, *args: PS.args, **kwargs: PS.kwargs
    ) -> Result[T_Return_cov, E]:
        pass


class T_Decorator(Protocol[PS, T_Return_contra, PS2, T_Return_cov2]):
    @abstractmethod
    def __call__(
        self, func: Callable[PS, T_Return_contra]
    ) -> Callable[PS2, T_Return_cov2]:
        pass


def ErrWithTraceback(e: Exception, extra_msg: str = "") -> Err[str]:
    if extra_msg:
        extra_msg = extra_msg.rstrip(" :\n")
        extra_msg = f"{extra_msg}"

    return Err(f"{extra_msg}\nException:\n{e}\n{traceback.format_exc()}")


def result_to_exitcode(r: Result[T, str], fail_code: int = 1) -> int:
    def _ok(_: T) -> int:
        return 0

    return r.map(_ok).unwrap_or(fail_code)


def exitcode_to_result(
    code: int, stdout: str = "", stderr: str = ""
) -> Result[str, str]:
    if code == 0:
        return Ok(f"Ok\n{stdout=}\n{stderr=}")
    else:
        return Err(f"Failure: exit code = {code}\n{stdout=}\n{stderr=}")


def result_reduce_list_separate(
    lst: Iterable[Result[T, str]]
) -> Tuple[list[T], list[str]]:
    oks: list[T] = []
    errs: list[str] = []
    for item in lst:
        match item:
            case Ok(x):
                oks.append(x)
            case Err(e):
                errs.append(e)

    return oks, errs


def result_reduce_dict_separate(
    dct: Mapping[K, Result[T, str]]
) -> Tuple[dict[K, T], list[str]]:
    oks: list[Tuple[K, T]] = []
    errs: list[str] = []
    for k, v in dct.items():
        match v:
            case Ok(v_):
                oks.append((k, v_))
            case Err(e):
                errs.append(e)

    return dict(oks), errs


async def result_reduce_list_separate_async(
    lst: Iterable[Awaitable[Result[T, str]]]
) -> Tuple[list[T], list[str]]:
    async def _get_result_value(awaitable: Awaitable[Result[T, str]]) -> Result[T, str]:
        return await awaitable

    values = await asyncio.gather(*[_get_result_value(a) for a in lst])
    return result_reduce_list_separate(values)


def result_reduce_list_all_ok(lst: Iterable[Result[T, str]]) -> Result[list[T], str]:
    oks, errs = result_reduce_list_separate(lst)
    if errs:
        return Err("\n".join(errs))
    else:
        return Ok(oks)


def result_reduce_dict_all_ok(
    dct: Mapping[K, Result[T, str]]
) -> Result[dict[K, T], str]:
    oks, errs = result_reduce_dict_separate(dct)
    if errs:
        return Err("\n".join(errs))
    else:
        return Ok(oks)


async def result_reduce_list_all_ok_async(
    lst: Iterable[Awaitable[Result[T, str]]]
) -> Result[list[T], str]:
    oks, errs = await result_reduce_list_separate_async(lst)
    if errs:
        return Err("\n".join(errs))
    else:
        return Ok(oks)


def parametrized(
    parametrized_decorator: Callable[
        Concatenate[Callable[PS2, T_Return], PS], Callable[PS3, T_Return2]
    ]
) -> Callable[PS, Callable[[Callable[PS2, T_Return]], Callable[PS3, T_Return2]]]:
    def make_decorator(
        *args: PS.args, **kwargs: PS.kwargs
    ) -> Callable[[Callable[PS2, T_Return]], Callable[PS3, T_Return2]]:
        def decorator(func: Callable[PS2, T_Return]) -> Callable[PS3, T_Return2]:
            return parametrized_decorator(func, *args, **kwargs)

        return decorator

    return make_decorator


@parametrized
def exception_handled(
    fn: T_UnsafeFn[PS, T_Return],
    exception_handler: Callable[[Exception], Result[T_Return, T_Err]],
) -> T_SafeFn[PS, T_Return, T_Err]:
    """
    Parametrized decorator for handling exceptions. 
    User defines an exception-handling function (`exception_handler`) that converts an Exception to a Result.
    
    Returns:
        A plain decorator, i.e. you can use it with @ as follows:
        ```
        @exception_handled(my_exception_handler)
        def my_raising_function(...):
            raise Exception("Something went wrong")
        ```
    """
    def decorated(*args: PS.args, **kwargs: PS.kwargs) -> Result[T_Return, T_Err]:
        try:
            return Ok(fn(*args, **kwargs))
        except Exception as e:
            return exception_handler(e)

    return decorated


@parametrized
def exception_handled_async(
    fn: T_AsyncUnsafeFn[PS, T_Return],
    exception_handler: Callable[[Exception], Result[T_Return, T_Err]],
) -> T_AsyncSafeFn[PS, T_Return, T_Err]:
    async def decorated(*args: PS.args, **kwargs: PS.kwargs) -> Result[T_Return, T_Err]:
        try:
            return Ok(await fn(*args, **kwargs))
        except Exception as e:
            return exception_handler(e)

    return decorated


def wrap_exception_traceback(e: Exception) -> Err[str]:
    return ErrWithTraceback(e)


def wrap_exception(e: Exception) -> Err[Exception]:
    return Err(e)


exception_to_err_with_traceback = exception_handled(wrap_exception_traceback)
exception_to_err = exception_handled(wrap_exception)
exception_to_err_with_traceback_async = exception_handled_async(
    wrap_exception_traceback
)
exception_to_err_async = exception_handled_async(wrap_exception)
